import pytest  # type: ignore

from ua_parser.utils import fa_simplifier


@pytest.mark.parametrize(
    ("from_", "to"),
    [
        (r"\d", "[0-9]"),
        (r"[\d]", "[0-9]"),
        (r"[\d\.]", r"[0-9\.]"),
    ],
)
def test_classes(from_, to):
    assert fa_simplifier(from_) == to



import io

from ua_parser import (
    BasicResolver,
    Domain,
    PartialResult,
    UserAgent,
)
from ua_parser.loaders import load_yaml
from ua_parser.matchers import UserAgentMatcher


def test_trivial_matching():
    p = BasicResolver(([UserAgentMatcher("(a)")], [], []))

    assert p("x", Domain.ALL) == PartialResult(
        string="x",
        domains=Domain.ALL,
        user_agent=None,
        os=None,
        device=None,
    )

    assert p("a", Domain.ALL) == PartialResult(
        string="a",
        domains=Domain.ALL,
        user_agent=UserAgent("a"),
        os=None,
        device=None,
    )


def test_partial():
    p = BasicResolver(([UserAgentMatcher("(a)")], [], []))

    assert p("x", Domain.USER_AGENT) == PartialResult(
        string="x",
        domains=Domain.USER_AGENT,
        user_agent=None,
        os=None,
        device=None,
    )

    assert p("a", Domain.USER_AGENT) == PartialResult(
        string="a",
        domains=Domain.USER_AGENT,
        user_agent=UserAgent("a"),
        os=None,
        device=None,
    )


def test_init_yaml():
    assert load_yaml
    f = io.BytesIO(
        b"""\
user_agent_parsers:
- regex: (a)
os_parsers: []
device_parsers: []
"""
    )
    p = BasicResolver(load_yaml(f))

    assert p("x", Domain.USER_AGENT) == PartialResult(
        string="x",
        domains=Domain.USER_AGENT,
        user_agent=None,
        os=None,
        device=None,
    )

    assert p("a", Domain.USER_AGENT) == PartialResult(
        string="a",
        domains=Domain.USER_AGENT,
        user_agent=UserAgent("a"),
        os=None,
        device=None,
    )



"""The package provides top-level helpers which use a lazily initialised
default parser. These are convenience functions, for more control it
is perfectly acceptable to instantiate and call parsers directly.

The default parser does use a cache keyed on the user-agent string,
but its exact behaviour is unspecified, if you require a consistent
behaviour or specific algorithm, set up your own parser (global or
not).

For convenience, direct aliases are also provided for:

- :mod:`core types <.types>`
- :mod:`caching utilities <.caching>`
- :mod:`ua_parser.basic.Parser` as :class:`BasicParser`

This way importing anything but the top-level package should not be
necessary unless you want to *implement* a parser.
"""

from __future__ import annotations

__all__ = [
    "OS",
    "BasicResolver",
    "Cache",
    "CachingResolver",
    "DefaultedResult",
    "Device",
    "Domain",
    "Matchers",
    "PartialResult",
    "Resolver",
    "Result",
    "UserAgent",
    "load_builtins",
    "load_lazy_builtins",
    "parse",
    "parse_device",
    "parse_os",
    "parse_user_agent",
]

import importlib.util
import threading
from typing import Callable, Optional, cast

from .basic import Resolver as BasicResolver
from .caching import CachingResolver, S3Fifo as Cache
from .core import (
    DefaultedResult,
    Device,
    Domain,
    Matchers,
    OS,
    PartialResult,
    Resolver,
    Result,
    UserAgent,
)
from .loaders import load_builtins, load_lazy_builtins
from .utils import IS_GRAAL

_ResolverCtor = Callable[[Matchers], Resolver]
Re2Resolver: Optional[_ResolverCtor] = None
if importlib.util.find_spec("re2"):
    from .re2 import Resolver as Re2Resolver
RegexResolver: Optional[_ResolverCtor] = None
if importlib.util.find_spec("ua_parser_rs"):
    from .regex import Resolver as RegexResolver
BestAvailableResolver: _ResolverCtor = next(
    filter(None, (RegexResolver, Re2Resolver, BasicResolver))
)


VERSION = (1, 0, 2)


class Parser:
    """Wrapper object, provides convenience methods around an
    underlying :class:`Resolver`.

    """

    @classmethod
    def from_matchers(cls, m: Matchers, /) -> Parser:
        """from_matchers(Matchers) -> Parser

        Instantiates a parser from the provided
        :class:`~ua_parser.core.Matchers` using the default resolver
        stack.

        """
        return cls(CachingResolver(BestAvailableResolver(m), Cache(2000)))

    def __init__(self, resolver: Resolver) -> None:
        self.resolver = resolver

    def __call__(self, ua: str, domains: Domain, /) -> PartialResult:
        """Parses the ``ua`` string, returning a parse result with *at least*
        the requested :class:`domains <Domain>` resolved (whether to success or
        failure).
        """
        return self.resolver(ua, domains)

    def parse(self: Resolver, ua: str) -> Result:
        """Convenience method for parsing all domains."""
        return self(ua, Domain.ALL).complete()

    def parse_user_agent(self: Resolver, ua: str) -> Optional[UserAgent]:
        """Convenience method for parsing the :class:`UserAgent` domain."""
        return self(ua, Domain.USER_AGENT).user_agent

    def parse_os(self: Resolver, ua: str) -> Optional[OS]:
        """Convenience method for parsing the :class:`OS` domain."""
        return self(ua, Domain.OS).os

    def parse_device(self: Resolver, ua: str) -> Optional[Device]:
        """Convenience method for parsing the :class:`Device` domain."""
        return self(ua, Domain.DEVICE).device


parser: Parser
"""Global :class:`Parser`, lazy-initialised on first access, used by
the global helper functions.

Can be *set* to configure a customised global parser.

Accessing the parser explicitely can be used eagerly force its
initialisation, rather than pay for it at first call.
"""

_lazy_globals_lock = threading.Lock()


def __getattr__(name: str) -> Parser:
    global parser
    with _lazy_globals_lock:
        if name == "parser":
            # if two threads access `ua_parser.parser` before it's
            # initialised, the second one will wait until the first
            # one's finished by which time the parser global should be
            # set and can be returned with no extra work
            if p := globals().get("parser"):
                return cast(Parser, p)

            if RegexResolver or Re2Resolver or IS_GRAAL:
                matchers = load_lazy_builtins()
            else:
                matchers = load_builtins()
            parser = Parser.from_matchers(matchers)
            return parser

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def parse(ua: str) -> Result:
    """Parses the :class:`.UserAgent`, :class:`.OS`, and :class:`.Device`
    information using the :data:`global parser <parser>`.

    Equivalent to calling each of :func:`parse_user_agent`,
    :func:`parse_os`, and :func:`parse_device` but *may* be more
    efficient than calling them separately depending on the underlying
    parser.

    Even in the best case, prefer the domain-specific helpers if
    you're not going to use *all* of them.
    """
    # import required to trigger __getattr__ and initialise the
    # parser, a `global` access fails to and we get a NameError
    from . import parser

    return parser(ua, Domain.ALL).complete()


def parse_user_agent(ua: str) -> Optional[UserAgent]:
    """Parses the :class:`browser <.UserAgent>` information using the
    :data:`global parser <parser>`.
    """
    from . import parser

    return parser(ua, Domain.USER_AGENT).user_agent


def parse_os(ua: str) -> Optional[OS]:
    """Parses the :class:`.OS` information using the :data:`global parser
    <parser>`.
    """
    from . import parser

    return parser(ua, Domain.OS).os


def parse_device(ua: str) -> Optional[Device]:
    """Parses the :class:`.Device` information using the :data:`global
    parser <parser>`.
    """
    from . import parser

    return parser(ua, Domain.DEVICE).device



from __future__ import annotations

__all__ = [
    "DeviceDict",
    "MatchersData",
    "OSDict",
    "UserAgentDict",
    "load_builtins",
    "load_data",
    "load_json",
    "load_lazy",
    "load_lazy_builtins",
    "load_yaml",
]

import io
import json
import os
from typing import (
    TYPE_CHECKING,
    Callable,
    List,
    Literal,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypedDict,
    Union,
    cast,
)

from . import lazy, matchers
from .core import Matchers

if TYPE_CHECKING:
    PathOrFile = Union[str, os.PathLike[str], io.IOBase]
    SafeLoader: Optional[Type[object]]
try:
    from yaml import CSafeLoader as SafeLoader, load
except ImportError:
    try:
        from yaml import SafeLoader, load
    except ImportError:
        load = SafeLoader = None  # type: ignore


def load_builtins() -> Matchers:
    """Loads the pre-compiled matcher data into eager matchers.

    The matchers data is imported lazily, and cached after imports,
    further imports simply reference the existing datas.

    """
    from ua_parser_builtins.matchers import MATCHERS

    # typing and mypy don't have safe upcast (#5756) and mypy is
    # unhappy about returning concrete matchers for a mixed type
    return cast(Matchers, MATCHERS)


def load_lazy_builtins() -> Matchers:
    """Loads the pre-compiled matcher data into lazy matchers.

    The matchers data is imported lazily, and cached after imports,
    further imports simply reference the existing datas.

    """
    from ua_parser_builtins.lazy import MATCHERS

    return cast(Matchers, MATCHERS)


# superclass needed to mix required & optional typed dict entries
# before 3.11 (and Required/NotRequired)
class _RegexDict(TypedDict):
    regex: str


class UserAgentDict(_RegexDict, total=False):
    family_replacement: str
    v1_replacement: str
    v2_replacement: str
    v3_replacement: str
    v4_replacement: str


class OSDict(_RegexDict, total=False):
    os_replacement: str
    os_v1_replacement: str
    os_v2_replacement: str
    os_v3_replacement: str
    os_v4_replacement: str


class DeviceDict(_RegexDict, total=False):
    regex_flag: Literal["i"]
    device_replacement: str
    brand_replacement: str
    model_replacement: str


MatchersData = Tuple[List[UserAgentDict], List[OSDict], List[DeviceDict]]
DataLoader = Callable[[MatchersData], Matchers]


def load_data(d: MatchersData) -> Matchers:
    """Loads the input data set into eager matchers."""
    return (
        [
            matchers.UserAgentMatcher(
                p["regex"],
                p.get("family_replacement"),
                p.get("v1_replacement"),
                p.get("v2_replacement"),
                p.get("v3_replacement"),
                p.get("v4_replacement"),
            )
            for p in d[0]
        ],
        [
            matchers.OSMatcher(
                p["regex"],
                p.get("os_replacement"),
                p.get("os_v1_replacement"),
                p.get("os_v2_replacement"),
                p.get("os_v3_replacement"),
                p.get("os_v4_replacement"),
            )
            for p in d[1]
        ],
        [
            matchers.DeviceMatcher(
                p["regex"],
                p.get("regex_flag"),
                p.get("device_replacement"),
                p.get("brand_replacement"),
                p.get("model_replacement"),
            )
            for p in d[2]
        ],
    )


def load_lazy(d: MatchersData) -> Matchers:
    """Loads the input data set into lazy matchers."""
    return (
        [
            lazy.UserAgentMatcher(
                p["regex"],
                p.get("family_replacement"),
                p.get("v1_replacement"),
                p.get("v2_replacement"),
                p.get("v3_replacement"),
                p.get("v4_replacement"),
            )
            for p in d[0]
        ],
        [
            lazy.OSMatcher(
                p["regex"],
                p.get("os_replacement"),
                p.get("os_v1_replacement"),
                p.get("os_v2_replacement"),
                p.get("os_v3_replacement"),
                p.get("os_v4_replacement"),
            )
            for p in d[1]
        ],
        [
            lazy.DeviceMatcher(
                p["regex"],
                p.get("regex_flag"),
                p.get("device_replacement"),
                p.get("brand_replacement"),
                p.get("model_replacement"),
            )
            for p in d[2]
        ],
    )


class FileLoader(Protocol):
    def __call__(
        self, path: PathOrFile, loader: DataLoader = load_data
    ) -> Matchers: ...


def load_json(f: PathOrFile, loader: DataLoader = load_data) -> Matchers:
    """Loads JSON data following the ``regexes.yaml`` structure.

    The ``loader`` parameter customises which matcher variant is
    generated, by default :func:`load_data` is used to generate eager
    matchers, :func:`load_lazy` can be used to generate lazy matchers
    instead.

    """
    if isinstance(f, (str, os.PathLike)):
        with open(f, "rb") as fp:
            regexes = json.load(fp)
    else:
        regexes = json.load(f)

    return loader(
        (
            regexes["user_agent_parsers"],
            regexes["os_parsers"],
            regexes["device_parsers"],
        )
    )


load_yaml: Optional[FileLoader]
if load is None:
    load_yaml = None
else:

    def load_yaml(path: PathOrFile, loader: DataLoader = load_data) -> Matchers:
        """Loads YAML data following the ``regexes.yaml`` structure.

        The ``loader`` parameter customises which matcher variant is
        generated, by default :func:`load_data` is used to generate eager
        matchers, :func:`load_lazy` cab be used to generate lazy matchers
        instead.
        """
        if isinstance(path, (str, os.PathLike)):
            with open(path, "rb") as fp:
                regexes = load(fp, Loader=SafeLoader)  # type: ignore
        else:
            regexes = load(path, Loader=SafeLoader)  # type: ignore

        return load_data(
            (
                regexes["user_agent_parsers"],
                regexes["os_parsers"],
                regexes["device_parsers"],
            )
        )



__all__ = ["DeviceMatcher", "OSMatcher", "UserAgentMatcher"]

import re
from typing import Literal, Optional, Pattern

from .core import Device, Matcher, OS, UserAgent
from .utils import get, replacer


class UserAgentMatcher(Matcher[UserAgent]):
    """Eager user agent matcher, compiles the input ``regex`` at
    initialisation.

    """

    pattern: Pattern[str]
    family: str
    major: Optional[str]
    minor: Optional[str]
    patch: Optional[str]
    patch_minor: Optional[str]

    def __init__(
        self,
        regex: str,
        family: Optional[str] = None,
        major: Optional[str] = None,
        minor: Optional[str] = None,
        patch: Optional[str] = None,
        patch_minor: Optional[str] = None,
    ) -> None:
        self.pattern = re.compile(regex)
        self.family = family or "$1"
        self.major = major
        self.minor = minor
        self.patch = patch
        self.patch_minor = patch_minor

    def __call__(self, ua: str) -> Optional[UserAgent]:
        if m := self.pattern.search(ua):
            return UserAgent(
                family=(
                    self.family.replace("$1", m[1])
                    if "$1" in self.family
                    else self.family
                ),
                major=self.major or get(m, 2),
                minor=self.minor or get(m, 3),
                patch=self.patch or get(m, 4),
                patch_minor=self.patch_minor or get(m, 5),
            )
        return None

    @property
    def regex(self) -> str:
        return self.pattern.pattern

    def __repr__(self) -> str:
        fields = [
            ("family", self.family if self.family != "$1" else None),
            ("major", self.major),
            ("minor", self.minor),
            ("patch", self.patch),
            ("patch_minor", self.patch_minor),
        ]
        args = "".join(f", {k}={v!r}" for k, v in fields if v is not None)

        return f"UserAgentMatcher({self.regex!r}{args})"


class OSMatcher(Matcher[OS]):
    """Eager OS matcher, compiles the input ``regex`` at
    initialisation.

    """

    pattern: Pattern[str]
    family: str
    major: str
    minor: str
    patch: str
    patch_minor: str

    def __init__(
        self,
        regex: str,
        family: Optional[str] = None,
        major: Optional[str] = None,
        minor: Optional[str] = None,
        patch: Optional[str] = None,
        patch_minor: Optional[str] = None,
    ) -> None:
        self.pattern = re.compile(regex)
        self.family = family or "$1"
        self.major = major or "$2"
        self.minor = minor or "$3"
        self.patch = patch or "$4"
        self.patch_minor = patch_minor or "$5"

    def __call__(self, ua: str) -> Optional[OS]:
        if m := self.pattern.search(ua):
            family = replacer(self.family, m)
            if family is None:
                raise ValueError(f"Unable to find OS family in {ua}")
            return OS(
                family=family,
                major=replacer(self.major, m),
                minor=replacer(self.minor, m),
                patch=replacer(self.patch, m),
                patch_minor=replacer(self.patch_minor, m),
            )
        return None

    @property
    def regex(self) -> str:
        return self.pattern.pattern

    def __repr__(self) -> str:
        fields = [
            ("family", self.family if self.family != "$1" else None),
            ("major", self.major if self.major != "$2" else None),
            ("minor", self.minor if self.minor != "$3" else None),
            ("patch", self.patch if self.patch != "$4" else None),
            ("patch_minor", self.patch_minor if self.patch_minor != "$5" else None),
        ]
        args = "".join(f", {k}={v!r}" for k, v in fields if v is not None)

        return f"OSMatcher({self.regex!r}{args})"


class DeviceMatcher(Matcher[Device]):
    """Eager device matcher, compiles the input ``regex`` at
    initialisation.

    """

    pattern: Pattern[str]
    family: str
    brand: str
    model: str

    def __init__(
        self,
        regex: str,
        regex_flag: Optional[Literal["i"]] = None,
        family: Optional[str] = None,
        brand: Optional[str] = None,
        model: Optional[str] = None,
    ) -> None:
        self.pattern = re.compile(
            regex, flags=re.IGNORECASE if regex_flag == "i" else 0
        )
        self.family = family or "$1"
        self.brand = brand or ""
        self.model = model or "$1"

    def __call__(self, ua: str) -> Optional[Device]:
        if m := self.pattern.search(ua):
            family = replacer(self.family, m)
            if family is None:
                raise ValueError(f"Unable to find device family in {ua}")
            return Device(
                family=family,
                brand=replacer(self.brand, m),
                model=replacer(self.model, m),
            )
        return None

    @property
    def regex(self) -> str:
        return self.pattern.pattern

    @property
    def regex_flag(self) -> str:
        return "i" if self.flags & re.IGNORECASE else ""

    @property
    def flags(self) -> int:
        return self.pattern.flags

    def __repr__(self) -> str:
        fields = [
            ("family", self.family if self.family != "$1" else None),
            ("brand", self.brand or None),
            ("model", self.model if self.model != "$1" else None),
        ]
        iflag = ', "i"' if self.flags & re.IGNORECASE else ""
        args = iflag + "".join(f", {k}={v!r}" for k, v in fields if v is not None)

        return f"DeviceMatcher({self.regex!r}{args})"



import ua_parser
from ua_parser import Domain, Parser, PartialResult, Result


def test_parser_memoized() -> None:
    """The global parser should be lazily instantiated but memoized"""
    # ensure there is no global parser
    vars(ua_parser).pop("parser", None)

    p1 = ua_parser.parser
    p2 = ua_parser.parser

    assert p1 is p2

    # force the creation of a clean parser
    del ua_parser.parser
    p3 = ua_parser.parser
    assert p3 is not p1


def resolver(s: str, d: Domain) -> PartialResult:
    return PartialResult(d, None, None, None, s)


def test_parser_utility() -> None:
    """Tests that ``Parser``'s methods to behave as procedural
    helpers, for users who may not wish to instantiate a parser or
    something.

    """

    r = Parser.parse(resolver, "a")
    assert r == Result(None, None, None, "a")

    os = Parser.parse_os(resolver, "a")
    assert os is None



import argparse
import datetime
import pathlib
import shutil
import subprocess

CORE_REMOTE = "https://github.com/ua-parser/uap-core"


parser = argparse.ArgumentParser(
    description="""Updates `uap-core` to `ref` and tags it with `version`

If successful, writes the commit to `REVISION` and prints it to stdout.
"""
)
parser.add_argument(
    "--ref",
    default="HEAD",
    help="uap-core ref to build, defaults to HEAD (the head of the default branch)",
)
parser.add_argument(
    "--version",
    help="version to tag the package as, defaults to an YMD calendar version matching the ref's commit date",
)
args = parser.parse_args()


if not shutil.which("git"):
    exit("git required")

r = subprocess.run(
    ["git", "ls-remote", CORE_REMOTE, args.ref],
    encoding="utf-8",
    stdout=subprocess.PIPE,
)
if r.returncode:
    exit("Unable to query uap-core repo")

if r.stdout:
    if r.stdout.count("\n") > 1:
        exit(f"Found multiple matching refs for {args.ref}:\n{r.stdout}")
    commit, _rest = r.stdout.split("\t", 1)
else:
    try:
        int(args.ref, 16)
        commit = args.ref
    except ValueError:
        exit(f"Unknown or invalid ref {args.ref!r}")

CORE_PATH = pathlib.Path(__file__).resolve().parent.parent / "uap-core"

r = subprocess.run(["git", "-C", CORE_PATH, "fetch", CORE_REMOTE, commit])
if r.returncode:
    exit(f"Unable to retrieve commit {commit!r}")

if args.version:
    tagname = args.version
else:
    r = subprocess.run(
        ["git", "-C", CORE_PATH, "show", "-s", "--format=%cs", commit],
        encoding="utf-8",
        stdout=subprocess.PIPE,
    )
    if r.returncode or not r.stdout:
        exit(f"Unable to retrieve commit date from commit {commit!r}")

    tagname = datetime.date.fromisoformat(r.stdout.rstrip()).strftime("%Y%m%d")

subprocess.run(["git", "-C", CORE_PATH, "switch", "-d", commit])
subprocess.run(["git", "-C", CORE_PATH, "tag", tagname, commit])
CORE_PATH.joinpath("REVISION").write_text(commit)
print(commit)



from __future__ import annotations

import io
import json
import os
import os.path
import tempfile
from contextlib import contextmanager
from typing import Any, Callable, ClassVar, Iterator, cast

import yaml
from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.metadata.plugin.interface import MetadataHookInterface
from versioningit import errors, get_version


class MetadataHook(MetadataHookInterface):
    def update(self, metadata: dict[str, Any]) -> None:
        try:
            v = get_version(
                os.path.join(self.root, "uap-core"),
                config={
                    "format": {
                        "distance": "{next_version}.dev{distance}",
                    }
                },
            )
        except errors.NotSdistError:
            with open(os.path.join(self.root, "uap-core", "package.json")) as ufile:
                ujson = json.load(ufile)
                v = ujson["version"]
        if v in ("0.15.0", "0.16.0", "0.18.0"):
            v = f"{v}.post1"

        metadata["version"] = v


class CompilerHook(BuildHookInterface):
    def initialize(
        self,
        version: str,
        build_data: dict[str, Any],
    ) -> None:
        rev = os.path.join(self.root, "uap-core/REVISION")
        if os.path.exists(rev):
            build_data["force_include"][rev] = "REVISION"

        with open(os.path.join(self.root, "uap-core/regexes.yaml"), "rb") as f:
            data = yaml.safe_load(f)

        with (
            tempfile.NamedTemporaryFile(delete=False) as matchers,
            tempfile.NamedTemporaryFile(delete=False) as lazy,
            tempfile.NamedTemporaryFile(delete=False) as regexes,
        ):
            matchers_w = EagerWriter(cast(io.RawIOBase, matchers))
            lazy_w = LazyWriter(cast(io.RawIOBase, lazy))
            legacy_w = LegacyWriter(cast(io.RawIOBase, regexes))

            for section, specs in data.items():
                with (
                    matchers_w.section(section),
                    lazy_w.section(section),
                    legacy_w.section(section),
                ):
                    extract = EXTRACTORS[section]
                    for s in specs:
                        el = trim(extract(s))
                        matchers_w.item(el)
                        lazy_w.item(el)
                        legacy_w.item(el)

            matchers_w.end()
            lazy_w.end()
            legacy_w.end()

        build_data["force_include"][matchers.name] = "ua_parser_builtins/matchers.py"
        build_data["force_include"][lazy.name] = "ua_parser_builtins/lazy.py"
        build_data["force_include"][regexes.name] = "ua_parser_builtins/regexes.py"

    def finalize(
        self,
        version: str,
        build_data: dict[str, Any],
        artifact_path: str,
    ):
        tempdir = tempfile.gettempdir()
        for k in build_data["force_include"]:
            if k.startswith(tempdir):
                os.remove(k)


def trim(items: list[str | None]) -> list[str | None]:
    """Removes trailing `None` from the extraction"""
    while len(items) > 1 and items[-1] is None:
        items.pop()
    return items


EXTRACTORS: dict[str, Callable[[dict[str, str]], list[str | None]]] = {
    "user_agent_parsers": lambda p: [
        p["regex"],
        p.get("family_replacement"),
        p.get("v1_replacement"),
        p.get("v2_replacement"),
        p.get("v3_replacement"),
        p.get("v4_replacement"),
    ],
    "os_parsers": lambda p: [
        p["regex"],
        p.get("os_replacement"),
        p.get("os_v1_replacement"),
        p.get("os_v2_replacement"),
        p.get("os_v3_replacement"),
        p.get("os_v4_replacement"),
    ],
    "device_parsers": lambda p: [
        p["regex"],
        p.get("regex_flag"),
        p.get("device_replacement"),
        p.get("brand_replacement"),
        p.get("model_replacement"),
    ],
}


class Writer:
    items: ClassVar[dict[str, bytes]]
    sections: ClassVar[dict[str, bytes]]
    prefix: bytes
    suffix = b""
    section_end = b""

    def __init__(self, fp: io.RawIOBase) -> None:
        self.fp = fp
        self.fp.write(
            b"""\
########################################################
# NOTICE: this file is autogenerated from regexes.yaml #
########################################################
"""
        )
        self.fp.write(self.prefix)
        self._section: str | None = None

    @contextmanager
    def section(self, id: str) -> Iterator[None]:
        self._section = id
        self.fp.write(self.sections[id])
        yield
        self.fp.write(self.section_end)

    def item(self, elements: list[str | None]) -> None:
        #        DeviceMatcher(re, flag, repl1),
        # assume we're in a section
        self.fp.write(self.items[cast(str, self._section)])
        self.fp.write(", ".join(map(repr, elements)).encode())
        self.fp.write(b"),\n")

    def end(self) -> None:
        self.fp.write(self.suffix)


class LegacyWriter(Writer):
    prefix = b"""\
__all__ = [
    "USER_AGENT_PARSERS",
    "DEVICE_PARSERS",
    "OS_PARSERS",
]

from ua_parser.user_agent_parser import UserAgentParser, DeviceParser, OSParser

"""
    sections: ClassVar[dict[str, bytes]] = {
        "user_agent_parsers": b"USER_AGENT_PARSERS = [\n",
        "os_parsers": b"\n\nOS_PARSERS = [\n",
        "device_parsers": b"\n\nDEVICE_PARSERS = [\n",
    }
    section_end = b"]"
    items: ClassVar[dict[str, bytes]] = {
        "user_agent_parsers": b"    UserAgentParser(",
        "os_parsers": b"    OSParser(",
        "device_parsers": b"    DeviceParser(",
    }
    suffix = b"\n"


class EagerWriter(Writer):
    prefix = b"""\
__all__ = ["MATCHERS"]

from typing import Tuple, List
from ua_parser.matchers import UserAgentMatcher, OSMatcher, DeviceMatcher

MATCHERS: Tuple[List[UserAgentMatcher], List[OSMatcher], List[DeviceMatcher]] = ([
"""
    sections: ClassVar[dict[str, bytes]] = {
        "user_agent_parsers": b"",
        "os_parsers": b"], [\n",
        "device_parsers": b"], [\n",
    }
    items: ClassVar[dict[str, bytes]] = {
        "user_agent_parsers": b"    UserAgentMatcher(",
        "os_parsers": b"    OSMatcher(",
        "device_parsers": b"    DeviceMatcher(",
    }
    suffix = b"])\n"


class LazyWriter(EagerWriter):
    prefix = b"""\
__all__ = ["MATCHERS"]

from typing import Tuple, List
from ua_parser.lazy import UserAgentMatcher, OSMatcher, DeviceMatcher

MATCHERS: Tuple[List[UserAgentMatcher], List[OSMatcher], List[DeviceMatcher]] = ([
"""



import abc
from dataclasses import dataclass
from enum import Flag, auto
from typing import Generic, List, Optional, Protocol, Tuple, TypeVar

__all__ = [
    "OS",
    "DefaultedResult",
    "Device",
    "Domain",
    "Matchers",
    "PartialResult",
    "Resolver",
    "Result",
    "UserAgent",
]


@dataclass(frozen=True, slots=True)
class UserAgent:
    """Browser ("user agent" aka the software responsible for the request)
    information parsed from the user agent string.
    """

    family: str = "Other"
    major: Optional[str] = None
    minor: Optional[str] = None
    patch: Optional[str] = None
    patch_minor: Optional[str] = None


@dataclass(frozen=True, slots=True)
class OS:
    """OS information parsed from the user agent string."""

    family: str = "Other"
    major: Optional[str] = None
    minor: Optional[str] = None
    patch: Optional[str] = None
    patch_minor: Optional[str] = None


@dataclass(frozen=True, slots=True)
class Device:
    """Device information parsed from the user agent string."""

    family: str = "Other"
    brand: Optional[str] = None
    model: Optional[str] = None


class Domain(Flag):
    """Hint for selecting which domains are requested when asking for a
    :class:`PartialResult`.
    """

    #: browser (user agent) domain
    USER_AGENT = auto()
    #: os domain
    OS = auto()
    #: device domain
    DEVICE = auto()
    #: shortcut for all three domains
    ALL = USER_AGENT | OS | DEVICE


@dataclass(frozen=True, slots=True)
class DefaultedResult:
    """Variant of :class:`Result` where attributes are set
    to a default value if their resolution failed.

    For all domains, the default value has ``family`` set to
    ``"Other"`` and every other attribute set to ``None``.
    """

    user_agent: UserAgent
    os: OS
    device: Device
    string: str


@dataclass(frozen=True, slots=True)
class Result:
    """Complete result.

    For each attribute (and domain), either the resolution was a
    success (a match was found) and the corresponding data is set, or
    it was a failure and the value is `None`.

    """

    user_agent: Optional[UserAgent]
    os: Optional[OS]
    device: Optional[Device]
    string: str

    def with_defaults(self) -> DefaultedResult:
        """Replaces every failed domain by its default value.

        Roughly matches pre-1.0 semantics, and can allow for more
        uniform handling by the client if they don't want or need the
        lookup failure information.

        """

        return DefaultedResult(
            user_agent=self.user_agent or UserAgent(),
            os=self.os or OS(),
            device=self.device or Device(),
            string=self.string,
        )


@dataclass(frozen=True, slots=True)
class PartialResult:
    """Potentially partial (incomplete) result.

    Domain fields (``user_agent``, ``os``, and ``device``) can be:

    - unset if not parsed yet
    - set to a parsing failure
    - set to a parsing success

    The ``domains`` flags specify which is which: if a :class:`Domain`
    flag is set, the corresponding attribute was looked up and is
    either ``None`` for a resolution failure (no match was found) or a
    value for a parsing success.

    If the flag is unset, the field has not been looked up yet, in
    which case it can be anything (but should usually be ``None``).

    """

    domains: Domain
    user_agent: Optional[UserAgent]
    os: Optional[OS]
    device: Optional[Device]
    string: str

    def complete(self) -> Result:
        """Requires that the result be fully resolved (every attribute is set,
        even if to a lookup failure).

        :raises ValueError: if the result is not fully resolved
        """
        if self.domains != Domain.ALL:
            raise ValueError("Only a result with all attributes set can be completed")

        return Result(
            user_agent=self.user_agent,
            os=self.os,
            device=self.device,
            string=self.string,
        )


class Resolver(Protocol):
    """Resolver()

    The resolver is the thin central abstraction of ua-parser, and
    used to compose various objects into the resolution stack which
    best fits the system's needs.

    A resolver is any callable which takes a string ``ua`` and a
    :class:`Domain`, and returns a :class:`PartialResult` with at
    least the requested domains marked as resolved (whether
    successfully or not).

    A resolver may resolve more domains than requested, but it needs
    to resolve at least the requested domains.

    See :class:`PartialResult` for more information about its
    working.

    """

    @abc.abstractmethod
    def __call__(self, ua: str, domain: Domain, /) -> PartialResult:
        """Resolves the ``ua``."""
        ...


T = TypeVar("T")


class Matcher(abc.ABC, Generic[T]):
    """A matcher is an individual pattern-rule, able to match a user
    agent string and in case of success extract the relevant data.

    Matchers need to expose their pattern for bulk resolvers.

    """

    @abc.abstractmethod
    def __call__(self, ua: str) -> Optional[T]:
        """Applies the matcher to an input."""
        ...

    @property
    @abc.abstractmethod
    def regex(self) -> str:
        """Returns the matcher's pattern."""
        ...

    @property
    def flags(self) -> int:
        """Returns the matcher's pattern flags (only
        :data:`re.IGNORECASE` is supported, and only for
        :class:`Matcher` [:class:`Device`])

        """
        return 0


Matchers = Tuple[
    List[Matcher[UserAgent]],
    List[Matcher[OS]],
    List[Matcher[Device]],
]



import argparse
import contextlib
import hashlib
import json
import re
import shutil
import sys
import tempfile
import zipfile
from urllib import parse, request

parser = argparse.ArgumentParser(
    description="Retrieves the revision for the latest release of ua-parser-builtins",
)
parser.add_argument(
    "--domain",
    default="pypi.org",
)
args = parser.parse_args()

url = parse.urlunsplit(("https", args.domain, "simple/ua-parser-builtins", "", ""))

print("checking", url, file=sys.stderr)
res = request.urlopen(
    request.Request(
        url,
        headers={
            "Accept": "application/vnd.pypi.simple.v1+json",
        },
    )
)
if res.status != 200:
    exit(f"Failed to retrieve project distributions: {res.status}")

distributions = json.load(res)
version, distribution = next(
    (v, d)
    for v, d in zip(
        reversed(distributions["versions"]), reversed(distributions["files"])
    )
    if not d["yanked"]
    if re.fullmatch(
        r"(\d+!)?\d+(\.\d+)*(\.post\d+)?",
        v,
        flags=re.ASCII,
    )
)
print("latest version:", version, file=sys.stderr)

res = request.urlopen(distribution["url"])
if res.status != 200:
    exit(f"Failed to retrieve wheel: {res.status}")

with tempfile.SpooledTemporaryFile(256 * 1024) as tf:
    shutil.copyfileobj(res, tf)
    for name, val in distribution["hashes"].items():
        tf.seek(0)
        d = hashlib.file_digest(tf, name).hexdigest()
        if d != val:
            exit(f"{name} mismatch: expected {val!r} got {d!r}")
    tf.seek(0)
    with zipfile.ZipFile(tf) as z:
        # if the REVISION file is not found then it's fine it's a
        # pre-calver release (hopefully) and that means we should cut
        # a calver one
        with contextlib.suppress(KeyError):
            print(z.read("REVISION").decode())



import os
import re
import warnings
from typing import Any, Dict, Optional, Tuple


class UserAgentParser(object):
    def __init__(
        self, pattern, family_replacement=None, v1_replacement=None, v2_replacement=None
    ):
        """Initialize UserAgentParser.

        Args:
          pattern: a regular expression string
          family_replacement: a string to override the matched family (optional)
          v1_replacement: a string to override the matched v1 (optional)
          v2_replacement: a string to override the matched v2 (optional)
        """
        self.user_agent_re = re.compile(pattern)
        self.family_replacement = family_replacement
        self.v1_replacement = v1_replacement
        self.v2_replacement = v2_replacement

    def Parse(
        self, user_agent_string: str
    ) -> Tuple[
        Optional[str],
        Optional[str],
        Optional[str],
        Optional[str],
    ]:
        family, v1, v2, v3 = None, None, None, None
        match = self.user_agent_re.search(user_agent_string)
        if match:
            if self.family_replacement:
                if re.search(r"\$1", self.family_replacement):
                    family = re.sub(r"\$1", match[1], self.family_replacement)
                else:
                    family = self.family_replacement
            else:
                family = match[1]

            if self.v1_replacement:
                v1 = self.v1_replacement
            elif match.lastindex and match.lastindex >= 2:
                v1 = match[2] or None

            if self.v2_replacement:
                v2 = self.v2_replacement
            elif match.lastindex and match.lastindex >= 3:
                v2 = match[3] or None

            if match.lastindex and match.lastindex >= 4:
                v3 = match[4] or None

        return family, v1, v2, v3


class OSParser(object):
    def __init__(
        self,
        pattern,
        os_replacement=None,
        os_v1_replacement=None,
        os_v2_replacement=None,
        os_v3_replacement=None,
        os_v4_replacement=None,
    ):
        """Initialize UserAgentParser.

        Args:
          pattern: a regular expression string
          os_replacement: a string to override the matched os (optional)
          os_v1_replacement: a string to override the matched v1 (optional)
          os_v2_replacement: a string to override the matched v2 (optional)
          os_v3_replacement: a string to override the matched v3 (optional)
          os_v4_replacement: a string to override the matched v4 (optional)
        """
        self.user_agent_re = re.compile(pattern)
        self.os_replacement = os_replacement
        self.os_v1_replacement = os_v1_replacement
        self.os_v2_replacement = os_v2_replacement
        self.os_v3_replacement = os_v3_replacement
        self.os_v4_replacement = os_v4_replacement

    def Parse(
        self, user_agent_string: str
    ) -> Tuple[
        Optional[str],
        Optional[str],
        Optional[str],
        Optional[str],
        Optional[str],
    ]:
        os, os_v1, os_v2, os_v3, os_v4 = None, None, None, None, None
        match = self.user_agent_re.search(user_agent_string)
        if match:
            if self.os_replacement:
                os = MultiReplace(self.os_replacement, match)
            elif match.lastindex:
                os = match[1]

            if self.os_v1_replacement:
                os_v1 = MultiReplace(self.os_v1_replacement, match)
            elif match.lastindex and match.lastindex >= 2:
                os_v1 = match[2]

            if self.os_v2_replacement:
                os_v2 = MultiReplace(self.os_v2_replacement, match)
            elif match.lastindex and match.lastindex >= 3:
                os_v2 = match[3]

            if self.os_v3_replacement:
                os_v3 = MultiReplace(self.os_v3_replacement, match)
            elif match.lastindex and match.lastindex >= 4:
                os_v3 = match[4]

            if self.os_v4_replacement:
                os_v4 = MultiReplace(self.os_v4_replacement, match)
            elif match.lastindex and match.lastindex >= 5:
                os_v4 = match[5]

        return os, os_v1, os_v2, os_v3, os_v4


def MultiReplace(string, match):
    def _repl(m):
        index = int(m[1]) - 1
        group = match.groups()
        if index < len(group):
            return group[index]
        return ""

    _string = re.sub(r"\$(\d)", _repl, string).strip()
    return _string or None


class DeviceParser(object):
    def __init__(
        self,
        pattern,
        regex_flag=None,
        device_replacement=None,
        brand_replacement=None,
        model_replacement=None,
    ):
        """Initialize UserAgentParser.

        Args:
          pattern: a regular expression string
          device_replacement: a string to override the matched device (optional)
        """
        self.user_agent_re = re.compile(
            pattern, re.IGNORECASE if regex_flag == "i" else 0
        )
        self.device_replacement = device_replacement
        self.brand_replacement = brand_replacement
        self.model_replacement = model_replacement

    def Parse(
        self, user_agent_string: str
    ) -> Tuple[
        Optional[str],
        Optional[str],
        Optional[str],
    ]:
        device, brand, model = None, None, None
        match = self.user_agent_re.search(user_agent_string)
        if match:
            if self.device_replacement:
                device = MultiReplace(self.device_replacement, match)
            else:
                device = match[1]

            if self.brand_replacement:
                brand = MultiReplace(self.brand_replacement, match)

            if self.model_replacement:
                model = MultiReplace(self.model_replacement, match)
            elif len(match.groups()) > 0:
                model = match[1]

        return device, brand, model


MAX_CACHE_SIZE = 200
_PARSE_CACHE: Dict[str, Dict[str, Any]] = {}


def _lookup(ua):
    if not isinstance(ua, str):
        raise TypeError(f"Expected user agent to be a string, got {ua!r}")

    entry = _PARSE_CACHE.get(ua)
    if entry is not None:
        return entry

    if len(_PARSE_CACHE) >= MAX_CACHE_SIZE:
        _PARSE_CACHE.clear()

    v = _PARSE_CACHE[ua] = {"string": ua}
    return v


def _cached(ua, key, fn):
    entry = _lookup(ua)
    r = entry.get(key)
    if not r:
        r = entry[key] = fn(ua)
    return r


def Parse(user_agent_string, **_jsParseBits):
    """Parse all the things
    Args:
      user_agent_string: the full user agent string
    Returns:
      A dictionary containing all parsed bits
    """
    if _jsParseBits:
        warnings.warn(
            "javascript overrides are not used anymore",
            category=DeprecationWarning,
            stacklevel=2,
        )

    entry = _lookup(user_agent_string)
    # entry is complete, return directly
    if len(entry) == 4:
        return entry

    # entry is partially or entirely empty
    if "user_agent" not in entry:
        entry["user_agent"] = _ParseUserAgent(user_agent_string)
    if "os" not in entry:
        entry["os"] = _ParseOS(user_agent_string)
    if "device" not in entry:
        entry["device"] = _ParseDevice(user_agent_string)

    return entry


def ParseUserAgent(user_agent_string, **_jsParseBits):
    """Parses the user-agent string for user agent (browser) info.
    Args:
      user_agent_string: The full user-agent string.
    Returns:
      A dictionary containing parsed bits.
    """
    if _jsParseBits:
        warnings.warn(
            "javascript overrides are not used anymore",
            category=DeprecationWarning,
            stacklevel=2,
        )
    return _cached(user_agent_string, "user_agent", _ParseUserAgent)


def _ParseUserAgent(user_agent_string):
    for uaParser in USER_AGENT_PARSERS:
        family, v1, v2, v3 = uaParser.Parse(user_agent_string)
        if family:
            break

    family = family or "Other"
    return {
        "family": family,
        "major": v1 or None,
        "minor": v2 or None,
        "patch": v3 or None,
    }


def ParseOS(user_agent_string, **_jsParseBits):
    """Parses the user-agent string for operating system info
    Args:
      user_agent_string: The full user-agent string.
    Returns:
      A dictionary containing parsed bits.
    """
    if _jsParseBits:
        warnings.warn(
            "javascript overrides are not used anymore",
            category=DeprecationWarning,
            stacklevel=2,
        )
    return _cached(user_agent_string, "os", _ParseOS)


def _ParseOS(user_agent_string):
    for osParser in OS_PARSERS:
        os, os_v1, os_v2, os_v3, os_v4 = osParser.Parse(user_agent_string)
        if os:
            break
    os = os or "Other"
    return {
        "family": os,
        "major": os_v1,
        "minor": os_v2,
        "patch": os_v3,
        "patch_minor": os_v4,
    }


def ParseDevice(user_agent_string, **_jsParseBits):
    """Parses the user-agent string for device info.
    Args:
        user_agent_string: The full user-agent string.
    Returns:
        A dictionary containing parsed bits.
    """
    if _jsParseBits:
        warnings.warn(
            "javascript overrides are not used anymore",
            category=DeprecationWarning,
            stacklevel=2,
        )
    return _cached(user_agent_string, "device", _ParseDevice)


def _ParseDevice(user_agent_string):
    for deviceParser in DEVICE_PARSERS:
        device, brand, model = deviceParser.Parse(user_agent_string)
        if device:
            break

    if device is None:
        device = "Other"

    return {"family": device, "brand": brand, "model": model}


def PrettyUserAgent(family, v1=None, v2=None, v3=None):
    """Pretty user agent string."""
    if v3:
        if v3[0].isdigit():
            return f"{family} {v1}.{v2}.{v3}"
        else:
            return f"{family} {v1}.{v2}{v3}"
    elif v2:
        return f"{family} {v1}.{v2}"
    elif v1:
        return f"{family} {v1}"
    return family


def PrettyOS(os, os_v1=None, os_v2=None, os_v3=None, os_v4=None):
    """Pretty os string."""
    if os_v4:
        return f"{os} {os_v1}.{os_v2}.{os_v3}.{os_v4}"
    if os_v3:
        if os_v3[0].isdigit():
            return f"{os} {os_v1}.{os_v2}.{os_v3}"
        else:
            return f"{os} {os_v1}.{os_v2}{os_v3}"
    elif os_v2:
        return f"{os} {os_v1}.{os_v2}"
    elif os_v1:
        return f"{os} {os_v1}"
    return os


def ParseWithJSOverrides(
    user_agent_string,
    js_user_agent_string=None,
    js_user_agent_family=None,
    js_user_agent_v1=None,
    js_user_agent_v2=None,
    js_user_agent_v3=None,
):
    warnings.warn(
        "Use Parse (or a specialised parser)", DeprecationWarning, stacklevel=2
    )

    # Override via JS properties.
    if js_user_agent_family is not None and js_user_agent_family != "":
        family = js_user_agent_family
        v1 = None
        v2 = None
        v3 = None
        if js_user_agent_v1 is not None:
            v1 = js_user_agent_v1
        if js_user_agent_v2 is not None:
            v2 = js_user_agent_v2
        if js_user_agent_v3 is not None:
            v3 = js_user_agent_v3
    else:
        for parser in USER_AGENT_PARSERS:
            family, v1, v2, v3 = parser.Parse(user_agent_string)
            if family:
                break

    # Override for Chrome Frame IFF Chrome is enabled.
    if (
        js_user_agent_string
        and js_user_agent_string.find("Chrome/") > -1
        and user_agent_string.find("chromeframe") > -1
    ):
        family = "Chrome Frame (%s %s)" % (family, v1)
        ua_dict = ParseUserAgent(js_user_agent_string)
        v1 = ua_dict["major"]
        v2 = ua_dict["minor"]
        v3 = ua_dict["patch"]

    return family or "Other", v1, v2, v3


def Pretty(family, v1=None, v2=None, v3=None):
    warnings.warn("Use PrettyUserAgent", DeprecationWarning, stacklevel=2)
    if v3:
        if v3[0].isdigit():
            return f"{family} {v1}.{v2}.{v3}"
        else:
            return f"{family} {v1}.{v2}{v3}"
    elif v2:
        return f"{family} {v1}.{v2}"
    elif v1:
        return f"{family} {v1}"
    return family


def GetFilters(
    user_agent_string,
    js_user_agent_string=None,
    js_user_agent_family=None,
    js_user_agent_v1=None,
    js_user_agent_v2=None,
    js_user_agent_v3=None,
):
    warnings.warn("No use case anymore", DeprecationWarning, stacklevel=2)

    filters = {}
    filterdict = {
        "js_user_agent_string": js_user_agent_string,
        "js_user_agent_family": js_user_agent_family,
        "js_user_agent_v1": js_user_agent_v1,
        "js_user_agent_v2": js_user_agent_v2,
        "js_user_agent_v3": js_user_agent_v3,
    }
    for key, value in filterdict.items():
        if value is not None and value != "":
            filters[key] = value
    return filters


# Build the list of user agent parsers from YAML
UA_PARSER_YAML = os.environ.get("UA_PARSER_YAML")
if UA_PARSER_YAML:
    # This will raise an ImportError if missing, obviously since it's no
    # longer a requirement
    import yaml

    try:
        # Try and use libyaml bindings if available since faster,
        # pyyaml doesn't do it by default (yaml/pyyaml#436)
        from yaml import CSafeLoader as SafeLoader
    except ImportError:
        from yaml import SafeLoader  #  type: ignore

    with open(UA_PARSER_YAML, "rb") as fp:
        regexes = yaml.load(fp, Loader=SafeLoader)

    USER_AGENT_PARSERS = []
    for _ua_parser in regexes["user_agent_parsers"]:
        _regex = _ua_parser["regex"]

        _family_replacement = _ua_parser.get("family_replacement")
        _v1_replacement = _ua_parser.get("v1_replacement")
        _v2_replacement = _ua_parser.get("v2_replacement")

        USER_AGENT_PARSERS.append(
            UserAgentParser(
                _regex, _family_replacement, _v1_replacement, _v2_replacement
            )
        )

    OS_PARSERS = []
    for _os_parser in regexes["os_parsers"]:
        _regex = _os_parser["regex"]

        _os_replacement = _os_parser.get("os_replacement")
        _os_v1_replacement = _os_parser.get("os_v1_replacement")
        _os_v2_replacement = _os_parser.get("os_v2_replacement")
        _os_v3_replacement = _os_parser.get("os_v3_replacement")
        _os_v4_replacement = _os_parser.get("os_v4_replacement")

        OS_PARSERS.append(
            OSParser(
                _regex,
                _os_replacement,
                _os_v1_replacement,
                _os_v2_replacement,
                _os_v3_replacement,
                _os_v4_replacement,
            )
        )

    DEVICE_PARSERS = []
    for _device_parser in regexes["device_parsers"]:
        _regex = _device_parser["regex"]

        _regex_flag = _device_parser.get("regex_flag")
        _device_replacement = _device_parser.get("device_replacement")
        _brand_replacement = _device_parser.get("brand_replacement")
        _model_replacement = _device_parser.get("model_replacement")

        DEVICE_PARSERS.append(
            DeviceParser(
                _regex,
                _regex_flag,
                _device_replacement,
                _brand_replacement,
                _model_replacement,
            )
        )

    # Clean our our temporary vars explicitly
    # so they can't be reused or imported
    del regexes
    del yaml
    del SafeLoader
else:
    # Just load our pre-compiled versions
    from ua_parser_builtins.regexes import (
        DEVICE_PARSERS,
        OS_PARSERS,
        USER_AGENT_PARSERS,
    )



from collections import OrderedDict

import pytest  # type: ignore

from ua_parser import (
    CachingResolver,
    Domain,
    Parser,
    PartialResult,
)
from ua_parser.caching import Lru, S3Fifo, Sieve


def test_lru():
    """Tests that the cache entries do get moved when accessed, and are
    popped LRU-first.
    """
    cache = Lru(2)
    p = Parser(
        CachingResolver(lambda s, d: PartialResult(d, None, None, None, s), cache)
    )

    p.parse("a")
    p.parse("b")

    assert cache.cache == OrderedDict(
        [
            ("a", PartialResult(Domain.ALL, None, None, None, "a")),
            ("b", PartialResult(Domain.ALL, None, None, None, "b")),
        ]
    )

    p.parse("a")
    p.parse("c")
    assert cache.cache == OrderedDict(
        [
            ("a", PartialResult(Domain.ALL, None, None, None, "a")),
            ("c", PartialResult(Domain.ALL, None, None, None, "c")),
        ]
    )


@pytest.mark.parametrize("cache", [Lru, S3Fifo, Sieve])
def test_backfill(cache):
    """Tests that caches handle partial parsing correctly, by updating
    the existing entry when new parts get parsed, without evicting
    still-fitting entries.
    """
    misses = 0

    def resolver(ua: str, domains: Domain, /) -> PartialResult:
        nonlocal misses
        misses += 1
        return PartialResult(domains, None, None, None, ua)

    p = Parser(CachingResolver(resolver, cache(10)))

    # fill the cache first, no need to hit the entries twice because
    # S3 waits until it needs space in the main cache before demotes
    # (or promotes) from the probationary cache.
    for s in map(str, range(9)):
        p.parse(s)
    assert misses == 9
    # add a partial entry
    p.parse_user_agent("a")
    # fill the partial entry, counts as a miss since it needs to
    # resolve the new bit
    p.parse_os("a")
    assert misses == 11

    misses = 0
    # check that the original entries are all hits
    for s in map(str, range(9)):
        p.parse(s)
    assert misses == 0



__all__ = ["Resolver"]

import re
from itertools import chain
from operator import methodcaller
from typing import Any, List

from .core import (
    Device,
    Domain,
    Matcher,
    Matchers,
    OS,
    PartialResult,
    UserAgent,
)
from .utils import IS_GRAAL, fa_simplifier


class Resolver:
    """A simple pure-python resolver based around trying a number of
    regular expressions in sequence for each domain, and returning a
    result when one matches.

    """

    user_agent_matchers: List[Matcher[UserAgent]]
    os_matchers: List[Matcher[OS]]
    device_matchers: List[Matcher[Device]]

    def __init__(
        self,
        matchers: Matchers,
    ) -> None:
        self.user_agent_matchers, self.os_matchers, self.device_matchers = matchers
        if IS_GRAAL:
            matcher: Any
            kind = next(
                (
                    "eager" if hasattr(type(m), "regex") else "lazy"
                    for m in chain.from_iterable(matchers)
                ),
                None,
            )
            if kind == "eager":
                for matcher in chain.from_iterable(matchers):
                    matcher.pattern = re.compile(
                        fa_simplifier(matcher.pattern.pattern),
                        flags=matcher.pattern.flags,
                    )
            elif kind == "lazy":
                for matcher in chain.from_iterable(matchers):
                    matcher.regex = fa_simplifier(matcher.pattern.pattern)

    def __call__(self, ua: str, domains: Domain, /) -> PartialResult:
        parse = methodcaller("__call__", ua)
        return PartialResult(
            domains=domains,
            string=ua,
            user_agent=(
                next(
                    filter(None, map(parse, self.user_agent_matchers)),
                    None,
                )
                if Domain.USER_AGENT in domains
                else None
            ),
            os=(
                next(
                    filter(None, map(parse, self.os_matchers)),
                    None,
                )
                if Domain.OS in domains
                else None
            ),
            device=(
                next(
                    filter(None, map(parse, self.device_matchers)),
                    None,
                )
                if Domain.DEVICE in domains
                else None
            ),
        )



import logging
import os
import platform

import pytest  # type: ignore
import yaml

if platform.python_implementation() == "PyPy":
    from yaml import SafeLoader
else:
    try:
        from yaml import CSafeLoader as SafeLoader  # type: ignore
    except ImportError:
        logging.getLogger(__name__).warning(
            "PyYaml C extension not available to run tests, this will result "
            "in dramatic tests slowdown."
        )
        from yaml import SafeLoader


from ua_parser import user_agent_parser

TEST_RESOURCES_DIR = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "../uap-core"
)


class TestParse:
    def testBrowserscopeStrings(self):
        self.runUserAgentTestsFromYAML(
            os.path.join(TEST_RESOURCES_DIR, "tests/test_ua.yaml")
        )

    def testBrowserscopeStringsOS(self):
        self.runOSTestsFromYAML(os.path.join(TEST_RESOURCES_DIR, "tests/test_os.yaml"))

    def testStringsOS(self):
        self.runOSTestsFromYAML(
            os.path.join(TEST_RESOURCES_DIR, "test_resources/additional_os_tests.yaml")
        )

    def testStringsDevice(self):
        self.runDeviceTestsFromYAML(
            os.path.join(TEST_RESOURCES_DIR, "tests/test_device.yaml")
        )

    def testMozillaStrings(self):
        self.runUserAgentTestsFromYAML(
            os.path.join(
                TEST_RESOURCES_DIR, "test_resources/firefox_user_agent_strings.yaml"
            )
        )

    # NOTE: The YAML file used here is one output by makePGTSComparisonYAML()
    # below, as opposed to the pgts_browser_list-orig.yaml file.  The -orig
    # file is by no means perfect, but identifies many browsers that we
    # classify as "Other".  This test itself is mostly useful to know when
    # somthing in UA parsing changes.  An effort should be made to try and
    # reconcile the differences between the two YAML files.
    def testPGTSStrings(self):
        self.runUserAgentTestsFromYAML(
            os.path.join(TEST_RESOURCES_DIR, "test_resources/pgts_browser_list.yaml")
        )

    def testParseAll(self):
        user_agent_string = "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.4; fr; rv:1.9.1.5) Gecko/20091102 Firefox/3.5.5,gzip(gfe),gzip(gfe)"
        expected = {
            "device": {"family": "Mac", "brand": "Apple", "model": "Mac"},
            "os": {
                "family": "Mac OS X",
                "major": "10",
                "minor": "4",
                "patch": None,
                "patch_minor": None,
            },
            "user_agent": {
                "family": "Firefox",
                "major": "3",
                "minor": "5",
                "patch": "5",
            },
            "string": user_agent_string,
        }

        result = user_agent_parser.Parse(user_agent_string)
        assert result == expected, "UA: {0}\n expected<{1}> != actual<{2}>".format(
            user_agent_string, expected, result
        )

    # Run a set of test cases from a YAML file
    def runUserAgentTestsFromYAML(self, file_name):
        yamlFile = open(os.path.join(TEST_RESOURCES_DIR, file_name))
        yamlContents = yaml.load(yamlFile, Loader=SafeLoader)
        yamlFile.close()

        for test_case in yamlContents["test_cases"]:
            # Inputs to Parse()
            user_agent_string = test_case["user_agent_string"]

            # The expected results
            expected = {
                "family": test_case["family"],
                "major": test_case["major"],
                "minor": test_case["minor"],
                "patch": test_case["patch"],
            }

            result = {}
            result = user_agent_parser.ParseUserAgent(user_agent_string)
            assert result == expected, (
                "UA: {0}\n expected<{1}, {2}, {3}, {4}> != actual<{5}, {6}, {7}, {8}>".format(
                    user_agent_string,
                    expected["family"],
                    expected["major"],
                    expected["minor"],
                    expected["patch"],
                    result["family"],
                    result["major"],
                    result["minor"],
                    result["patch"],
                )
            )
            assert (
                len(user_agent_parser._PARSE_CACHE) <= user_agent_parser.MAX_CACHE_SIZE
            ), "verify that the cache size never exceeds the configured setting"

    def runOSTestsFromYAML(self, file_name):
        yamlFile = open(os.path.join(TEST_RESOURCES_DIR, file_name))
        yamlContents = yaml.load(yamlFile, Loader=SafeLoader)
        yamlFile.close()

        for test_case in yamlContents["test_cases"]:
            # Inputs to Parse()
            user_agent_string = test_case["user_agent_string"]

            # The expected results
            expected = {
                "family": test_case["family"],
                "major": test_case["major"],
                "minor": test_case["minor"],
                "patch": test_case["patch"],
                "patch_minor": test_case["patch_minor"],
            }

            result = user_agent_parser.ParseOS(user_agent_string)
            assert result == expected, (
                "UA: {0}\n expected<{1} {2} {3} {4} {5}> != actual<{6} {7} {8} {9} {10}>".format(
                    user_agent_string,
                    expected["family"],
                    expected["major"],
                    expected["minor"],
                    expected["patch"],
                    expected["patch_minor"],
                    result["family"],
                    result["major"],
                    result["minor"],
                    result["patch"],
                    result["patch_minor"],
                )
            )

    def runDeviceTestsFromYAML(self, file_name):
        yamlFile = open(os.path.join(TEST_RESOURCES_DIR, file_name))
        yamlContents = yaml.load(yamlFile, Loader=SafeLoader)
        yamlFile.close()

        for test_case in yamlContents["test_cases"]:
            # Inputs to Parse()
            user_agent_string = test_case["user_agent_string"]

            # The expected results
            expected = {
                "family": test_case["family"],
                "brand": test_case["brand"],
                "model": test_case["model"],
            }

            result = user_agent_parser.ParseDevice(user_agent_string)
            assert result == expected, (
                "UA: {0}\n expected<{1} {2} {3}> != actual<{4} {5} {6}>".format(
                    user_agent_string,
                    expected["family"],
                    expected["brand"],
                    expected["model"],
                    result["family"],
                    result["brand"],
                    result["model"],
                )
            )


class TestGetFilters:
    def testGetFiltersNoMatchesGiveEmptyDict(self):
        user_agent_string = "foo"
        with pytest.warns(DeprecationWarning):
            filters = user_agent_parser.GetFilters(
                user_agent_string, js_user_agent_string=None
            )
        assert {} == filters

    def testGetFiltersJsUaPassedThrough(self):
        user_agent_string = "foo"
        with pytest.warns(DeprecationWarning):
            filters = user_agent_parser.GetFilters(
                user_agent_string, js_user_agent_string="bar"
            )
        assert {"js_user_agent_string": "bar"} == filters

    def testGetFiltersJsUserAgentFamilyAndVersions(self):
        user_agent_string = (
            "Mozilla/4.0 (compatible; MSIE 8.0; "
            "Windows NT 5.1; Trident/4.0; GTB6; .NET CLR 2.0.50727; "
            ".NET CLR 3.0.4506.2152; .NET CLR 3.5.30729)"
        )
        with pytest.warns(DeprecationWarning):
            filters = user_agent_parser.GetFilters(
                user_agent_string,
                js_user_agent_string="bar",
                js_user_agent_family="foo",
            )
        assert {"js_user_agent_string": "bar", "js_user_agent_family": "foo"} == filters


class TestDeprecationWarnings:
    def test_parser_deprecation(self):
        with pytest.warns(DeprecationWarning) as ws:
            user_agent_parser.ParseWithJSOverrides("")
        assert len(ws) == 1

    def test_printer_deprecation(self):
        with pytest.warns(DeprecationWarning) as ws:
            user_agent_parser.Pretty("")
        assert len(ws) == 1

    def test_js_bits_deprecation(self):
        for parser, count in [
            (user_agent_parser.Parse, 1),
            (user_agent_parser.ParseUserAgent, 1),
            (user_agent_parser.ParseOS, 1),
            (user_agent_parser.ParseDevice, 1),
        ]:
            user_agent_parser._PARSE_CACHE.clear()
            with pytest.warns(DeprecationWarning) as ws:
                parser("some random thing", js_attribute=True)
            assert len(ws) == count



__all__ = ["Resolver"]

import re
from typing import List

import re2  # type: ignore

from .core import (
    Device,
    Domain,
    Matcher,
    Matchers,
    OS,
    PartialResult,
    UserAgent,
)
from .utils import fa_simplifier


class DummyFilter:
    def Match(self, _: str) -> None:
        pass


class Resolver:
    ua: re2.Filter
    user_agent_matchers: List[Matcher[UserAgent]]
    os: re2.Filter
    os_matchers: List[Matcher[OS]]
    devices: re2.Filter
    device_matchers: List[Matcher[Device]]

    def __init__(
        self,
        matchers: Matchers,
    ) -> None:
        self.user_agent_matchers, self.os_matchers, self.device_matchers = matchers

        if self.user_agent_matchers:
            self.ua = re2.Filter()
            for u in self.user_agent_matchers:
                self.ua.Add(fa_simplifier(u.regex))
            self.ua.Compile()
        else:
            self.ua = DummyFilter()

        if self.os_matchers:
            self.os = re2.Filter()
            for o in self.os_matchers:
                self.os.Add(fa_simplifier(o.regex))
            self.os.Compile()
        else:
            self.os = DummyFilter()

        if self.device_matchers:
            self.devices = re2.Filter()
            for d in self.device_matchers:
                # Prepend the i global flag if IGNORECASE is set. Assumes
                # no pattern uses global flags, but since they're not
                # supported in JS that seems safe.
                if d.flags & re.IGNORECASE:
                    self.devices.Add("(?i)" + fa_simplifier(d.regex))
                else:
                    self.devices.Add(fa_simplifier(d.regex))
            self.devices.Compile()
        else:
            self.devices = DummyFilter()

    def __call__(self, ua: str, domains: Domain, /) -> PartialResult:
        user_agent = os = device = None
        if Domain.USER_AGENT in domains:
            if matches := self.ua.Match(ua):
                # Set/Filter does not return the match in index order
                # (position order?) so to fit UAP semantics we need to
                # extract the first matching regex (lowest index).
                user_agent = self.user_agent_matchers[min(matches)](ua)
        if Domain.OS in domains:
            if matches := self.os.Match(ua):
                os = self.os_matchers[min(matches)](ua)
        if Domain.DEVICE in domains:
            if matches := self.devices.Match(ua):
                device = self.device_matchers[min(matches)](ua)
        return PartialResult(
            domains=domains, string=ua, user_agent=user_agent, os=os, device=device
        )



import argparse
import bisect
import collections
import csv
import gc
import io
import itertools
import math
import os
import random
import sys
import threading
import time
import types
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from . import (
    BasicResolver,
    CachingResolver,
    Domain,
    Matchers,
    Parser,
    PartialResult,
    Resolver,
    caching,
)
from .caching import Cache, Local
from .loaders import load_builtins, load_yaml

try:
    from .re2 import Resolver as Re2Resolver
except ImportError:
    pass
try:
    from .regex import Resolver as RegexResolver
except ImportError:
    pass
from .user_agent_parser import Parse

CACHEABLE = {
    "basic": True,
    "re2": True,
    "regex": True,
    "legacy": False,
}


CACHES: Dict[str, Optional[Callable[[int], Cache]]] = {"none": None}
CACHES.update(
    (cache.__name__.lower(), cache)
    for cache in [
        cast(Callable[[int], Cache], caching.Lru),
        caching.S3Fifo,
        caching.Sieve,
    ]
)

try:
    import tracemalloc
except ImportError:
    snapshot = types.SimpleNamespace(
        compare_to=lambda _1, _2: [],
    )
    tracemalloc = types.SimpleNamespace(  # type: ignore
        start=lambda: None,
        take_snapshot=lambda: snapshot,
    )


def get_rules(parsers: List[str], regexes: Optional[io.IOBase]) -> Matchers:
    if regexes:
        if not load_yaml:
            sys.exit("yaml loading unavailable, please install pyyaml")

        rules = load_yaml(regexes)
        if "legacy" in parsers:
            print(
                "The legacy parser is incompatible with custom regexes, ignoring.",
                file=sys.stderr,
            )
            parsers.remove("legacy")
    else:
        rules = load_builtins()

    return rules


def parse_item(item: str, all: list[str] | None) -> list[str]:
    if item == "*":
        assert all
        return all
    elif item.startswith("{"):
        assert item.endswith("}")
        return item[1:-1].split(",")
    else:
        return [item]


def rules_to_parsers(args: argparse.Namespace) -> Iterator[tuple[str, str, int]]:
    seen = set()
    for selector in args.selector:
        p, c, s = selector.split(":")
        for triplet in (
            (pp, "none" if ss == 0 else cc, ss)
            for pp in parse_item(p, ["basic", "re2", "regex", "legacy"])
            for cc in (parse_item(c, list(CACHES)) if CACHEABLE[pp] else ["none"])
            for ss in (map(int, parse_item(s, None)) if cc != "none" else [0])
        ):
            if triplet not in seen:
                seen.add(triplet)
                yield triplet


def run_stdout(args: argparse.Namespace) -> None:
    lines = list(map(sys.intern, args.file))
    count = len(lines)
    uniques = len(set(lines))
    print(f"{args.file.name}: {count} lines, {uniques} unique ({uniques / count:.0%})")

    parsers = list(rules_to_parsers(args))

    rules = get_rules([*{p for p, _, _ in parsers}], args.regexes)

    w = max(
        math.ceil(3 + len(p) + len(c) + (s and math.log10(s))) for p, c, s in parsers
    )
    for p, c, n in parsers:
        name = "-".join(map(str, filter(None, (p, c != "none" and c, n))))
        print(f"{name:{w}}", end=": ", flush=True)

        parser = get_parser(p, c, n, rules)
        t = run(parser, lines)

        secs = t / 1e9
        tpl = t / 1000 / len(lines)

        print(f"{secs:>5.2f}s ({tpl:>4.0f}us/line)")


def run_csv(args: argparse.Namespace) -> None:
    lines = list(map(sys.intern, args.file))
    LEN = len(lines) * 1000

    parsers = list(rules_to_parsers(args))
    if not parsers:
        sys.exit("No parser selected")

    rules = get_rules([*{p for p, _, _ in parsers}], args.regexes)
    columns = {"size": ""}
    columns.update(
        (f"{p}-{c}", p if c == "none" else f"{p}-{c}") for p, c, _ in parsers
    )
    w = csv.DictWriter(
        sys.stdout,
        list(columns),
        dialect="unix",
        quoting=csv.QUOTE_MINIMAL,
    )
    w.writerow(columns)

    parsers.sort(key=lambda t: t[2])
    grouped = itertools.groupby(parsers, key=lambda t: t[2])

    # these are the "template rows", which contain the no-cache
    # runs which get replicated on every cachesize row
    zeroes = {}
    # if we have entries with no cache size, compute them first so
    # we can apply them to every cachesize
    if parsers[0][2] == 0:
        (_, ps) = next(grouped)
        # cache could be ignored as it should always be `"none"`
        for parser, cache, _ in ps:
            p = get_parser(parser, cache, 0, rules)
            zeroes[f"{parser}-{cache}"] = run(p, lines) // LEN

    # special cases for configurations where we can't have
    # cachesize lines, write the template row out directly
    if (
        all(p == "legacy" for p, _, _ in parsers)
        or all(c == "none" for _, c, _ in parsers)
        or all(s == 0 for _, _, s in parsers)
    ):
        zeroes["size"] = 0
        w.writerow(zeroes)
        return

    for cachesize, ps in grouped:
        row = dict(zeroes, size=cachesize)
        for parser, cache, _ in ps:
            p = get_parser(parser, cache, cachesize, rules)
            row[f"{parser}-{cache}"] = run(p, lines) // LEN
        w.writerow(row)


def get_parser(
    parser: str, cache: str, cachesize: int, rules: Matchers
) -> Callable[[str], Any]:
    r: Resolver
    if parser == "legacy":
        return Parse
    elif parser == "basic":
        r = BasicResolver(rules)
    elif parser == "re2":
        r = Re2Resolver(rules)
    elif parser == "regex":
        r = RegexResolver(rules)
    else:
        sys.exit(f"unknown parser {parser!r}")

    if cache not in CACHES:
        sys.exit(f"unknown cache algorithm {cache!r}")

    c = CACHES.get(cache)
    if c is None:
        return Parser(r).parse

    return Parser(CachingResolver(r, c(cachesize))).parse


def run(
    parse: Callable[[str], None],
    lines: Iterable[str],
) -> int:
    t = time.perf_counter_ns()
    for line in lines:
        parse(line)
    return time.perf_counter_ns() - t


class Belady:
    def __init__(self, maxsize: int, data: List[str]):
        self.maxsize = maxsize
        self.cache: Dict[str, PartialResult] = {}
        self.queue: Deque[Tuple[int, str]] = collections.deque()
        self.distances: Dict[str, List[int]] = {}
        for i, e in enumerate(data):
            self.distances.setdefault(e, []).append(i)
        for freqs in self.distances.values():
            freqs.reverse()

    def __getitem__(self, key: str) -> Optional[PartialResult]:
        self.distances[key].pop()
        if c := self.cache.get(key):
            # on cache hit, the entry should be the lowest in the
            # queue
            assert self.queue.popleft()[1] == key
            # if the key has future occurrences
            if ds := self.distances[key]:
                # reinsert in queue
                bisect.insort(self.queue, (ds[-1], key))
            else:
                # otherwise remove from cache & occurrences map
                del self.cache[key]

        return c

    def __setitem__(self, key: str, entry: PartialResult) -> None:
        # if there are no future occurrences just bail
        ds = self.distances[key]
        if not ds:
            return

        next_distance = ds[-1]
        # if the cache has room, just add the entry
        if len(self.cache) >= self.maxsize:
            # if the next occurrence of the new entry is later than
            # every existing occurrence, ignore it
            if next_distance > self.queue[-1][0]:
                return
            # otherwise remove the latest entry
            _, k = self.queue.pop()
            del self.cache[k]

        self.cache[key] = entry
        bisect.insort(self.queue, (next_distance, key))


def run_hitrates(args: argparse.Namespace) -> None:
    r = PartialResult(
        domains=Domain.ALL,
        string="",
        user_agent=None,
        os=None,
        device=None,
    )

    class Counter:
        def __init__(self) -> None:
            self.count = 0

        def __call__(self, ua: str, domains: Domain, /) -> PartialResult:
            self.count += 1
            return r

    lines = list(map(sys.intern, args.file))
    total = len(lines)
    uniques = len(set(lines))
    print(total, "lines", uniques, "uniques")
    print()
    w = int(math.log10(max(args.cachesizes)) + 1)

    def belady(maxsize: int) -> Cache:
        return Belady(maxsize, lines)

    tracemalloc.start()
    for cache, cache_size in itertools.product(
        itertools.chain([belady], filter(None, CACHES.values())),
        args.cachesizes,
    ):
        misses = Counter()
        gc.collect()
        before = tracemalloc.take_snapshot()
        parser = Parser(CachingResolver(misses, cache(cache_size)))
        for line in lines:
            parser.parse(line)
        gc.collect()
        after = tracemalloc.take_snapshot()
        if cache == belady:
            diff = "{0:>14} {0:>12}".format("-")
        else:
            overhead = sum(s.size_diff for s in after.compare_to(before, "filename"))
            diff = "{:8} bytes ({:3.0f}b/entry)".format(
                overhead,
                overhead / cache_size,
            )
        print(
            f"{cache.__name__.lower():8}({cache_size:{w}}): {(total - misses.count) / total * 100:2.0f}% hit rate {diff}"
        )
        del misses, parser


CACHESIZE = 1000


def worker(
    start: threading.Event,
    parser: Parser,
    lines: Iterable[str],
    end: threading.Barrier,
) -> None:
    start.wait()

    for ua in lines:
        parser.parse(ua)

    end.wait()


def run_threaded(args: argparse.Namespace) -> None:
    lines = list(map(sys.intern, args.file))
    basic = BasicResolver(load_builtins())
    resolvers: List[Tuple[str, Resolver]] = [
        ("locking-lru", CachingResolver(basic, caching.Lru(CACHESIZE))),
        ("local-lru", CachingResolver(basic, Local(lambda: caching.Lru(CACHESIZE)))),
        ("re2", Re2Resolver(load_builtins())),
        ("regex", RegexResolver(load_builtins())),
    ]
    for name, resolver in resolvers:
        print(f"{name:11}: ", end="", flush=True)
        # randomize the dataset for each thread, predictably, to
        # simulate distributed load (not great but better than
        # nothing, and probably better than reusing the exact same
        # load)
        r = random.Random(42)
        start = threading.Event()
        end = threading.Barrier(args.threads + 1)

        parser = Parser(resolver)
        for _ in range(args.threads):
            threading.Thread(
                target=worker,
                args=(start, parser, r.sample(lines, len(lines)), end),
                daemon=True,
            ).start()

        st = time.perf_counter_ns()
        start.set()
        end.wait()

        # each thread gets len(lines), so total number of processed
        # lines is t*len(lines)
        totlines = len(lines) * args.threads
        # runtime in us
        t = (time.perf_counter_ns() - st) / 1000
        print(f"{t / totlines:>4.0f}us/line", flush=True)


EPILOG = """For good results the sample `file` should be an actual
non-sorted non-deduplicated sample of user agent strings from traffic
on a comparable (or the actual) site or application targeted for
classification."""

parser = argparse.ArgumentParser(prog="ua_parser", epilog="epi")
parser.set_defaults(func=None)

fp = argparse.ArgumentParser(add_help=False)
fp.add_argument(
    "file",
    type=argparse.FileType("r", encoding="utf-8"),
    help="Sample user agent file, the file must contain a single user agent "
    "string per line, use `-` for stdin.",
)

sub = parser.add_subparsers(title="commands")

bench = sub.add_parser(
    "bench",
    help="benchmark various parser configurations on sample files",
    parents=[fp],
    epilog=EPILOG,
    description="""Different sites and applications can have different
    traffic pattenrs, and thus want different setups and tradeoffs.
    This subcommand allows testing ua-parser's different base
    resolvers, caches, anc cache sizes in order to customise the
    parser to the application's requirements. It's also useful to
    bench the library itself though.""",
)
bench.add_argument(
    "-R",
    "--regexes",
    type=argparse.FileType("rb"),
    help="""Custom regexes.yaml file, if ommitted the benchmark will
    use the embedded regexes file rom uap-core. Custom regexes files
    can allow evaluating the performance impact of new rules or
    cut-down reference files (if legacy rules are nor relevant to your
    needs). Because YAML is (mostly) a superset of JSON, JSON regexes
    files will also work fine.""",
)


class ToFunc(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Union[str, Sequence[str], None],
        option_string: Optional[str] = None,
    ) -> None:
        if values == "stdout":
            setattr(namespace, self.dest, run_stdout)
        elif values == "csv":
            setattr(namespace, self.dest, run_csv)
        else:
            raise ValueError(f"invalid output {values!r}")


bench.add_argument(
    "-O",
    "--output",
    choices=["stdout", "csv"],
    default=run_stdout,
    dest="func",
    action=ToFunc,
    help="""By default (`stdout`) the result of each configuration /
    combination is printed to stdout with the combination name
    followed by the total parse time for the file and the per-entry
    average. `csv` will instead output a valid CSV table to stdout,
    with a parser combination per column and a cache size per row.
    Combinations without cache will have the same value on every row.
    If no combination uses a cache, the output will have a single row
    with a first cell of value 0.""",
)
bench.add_argument(
    "selector",
    nargs="*",
    default=["*:*:{10,20,50,100,200,500,1000,2000,5000}"],
    help=f"""A generative selector expression, composed of 3 parts: 1.
the parser (base), 2. the cache implementation ({", ".join(CACHES)})
and 3. the cache size. For parser and cache `*` is an alias for stands
in for "every value", a bracketed expression for an enumeration, and
the selector can be repeated to explicitly list each configuration """,
)

hitrates = sub.add_parser(
    "hitrates",
    help="measure hitrates of cache configurations against sample files",
    parents=[fp],
    epilog=EPILOG,
)
hitrates.set_defaults(func=run_hitrates)
hitrates.add_argument(
    "--cachesizes",
    nargs="+",
    type=int,
    default=[10, 20, 50, 100, 200, 500, 1000, 2000, 5000],
    help="""List of cache sizes to test hitrates for, for each cache
    algorithm. """,
)

threaded = sub.add_parser(
    "threading",
    help="estimate impact of concurrency and contention on different parser configurations",
    parents=[fp],
    epilog=EPILOG,
)
threaded.set_defaults(func=run_threaded)
threaded.add_argument(
    "-n",
    "--threads",
    type=int,
    default=os.cpu_count() or 1,
)

args = parser.parse_args()
if args.func:
    args.func(args)
else:
    parser.print_help()



__all__ = ["Resolver"]

from operator import attrgetter

import ua_parser_rs  # type: ignore

from .core import (
    Device,
    Domain,
    Matchers,
    OS,
    PartialResult,
    UserAgent,
)


class Resolver:
    ua: ua_parser_rs.UserAgentExtractor
    os: ua_parser_rs.OSExtractor
    de: ua_parser_rs.DeviceExtractor

    def __init__(self, matchers: Matchers) -> None:
        ua, os, de = matchers
        self.ua = ua_parser_rs.UserAgentExtractor(
            map(
                attrgetter("regex", "family", "major", "minor", "patch", "patch_minor"),
                ua,
            )
        )
        self.os = ua_parser_rs.OSExtractor(
            map(
                attrgetter("regex", "family", "major", "minor", "patch", "patch_minor"),
                os,
            )
        )
        self.de = ua_parser_rs.DeviceExtractor(
            map(
                attrgetter("regex", "regex_flag", "family", "brand", "model"),
                de,
            )
        )

    def __call__(self, ua: str, domains: Domain, /) -> PartialResult:
        user_agent = os = device = None
        if Domain.USER_AGENT in domains:
            if uam := self.ua.extract(ua):
                user_agent = UserAgent(
                    uam.family,
                    uam.major,
                    uam.minor,
                    uam.patch,
                    uam.patch_minor,
                )
        if Domain.OS in domains:
            if osm := self.os.extract(ua):
                os = OS(
                    osm.family,
                    osm.major,
                    osm.minor,
                    osm.patch,
                    osm.patch_minor,
                )
        if Domain.DEVICE in domains:
            if dem := self.de.extract(ua):
                device = Device(
                    dem.family,
                    dem.brand,
                    dem.model,
                )
        return PartialResult(
            domains=domains,
            string=ua,
            user_agent=user_agent,
            os=os,
            device=device,
        )



"""Tests UAP-Python using the UAP-core test suite"""

import dataclasses
import logging
import pathlib
from operator import attrgetter
from typing import cast

import pytest  # type: ignore

try:
    from yaml import (
        CSafeLoader as SafeLoader,
        load,
    )
except ImportError:
    logging.getLogger(__name__).warning(
        "PyYaml C extension not available to run tests, this will result "
        "in tests slowdown."
    )
    from yaml import SafeLoader, load  # type: ignore

from ua_parser import (
    BasicResolver,
    Device,
    OS,
    Parser,
    Result,
    UserAgent,
    load_builtins,
    load_lazy_builtins,
    loaders,
)
from ua_parser.matchers import UserAgentMatcher

CORE_DIR = (pathlib.Path(__name__).parent.parent / "uap-core").resolve()


data = cast(loaders.FileLoader, loaders.load_yaml)(CORE_DIR / "regexes.yaml")
data_lazy = cast(loaders.FileLoader, loaders.load_yaml)(
    CORE_DIR / "regexes.yaml", loader=loaders.load_lazy
)
PARSERS = [
    pytest.param(Parser(BasicResolver(load_builtins())), id="basic"),
    pytest.param(Parser(BasicResolver(load_lazy_builtins())), id="lazy"),
    pytest.param(Parser(BasicResolver(data)), id="basic-yaml"),
    pytest.param(Parser(BasicResolver(data_lazy)), id="lazy-yaml"),
]
try:
    from ua_parser import re2
except ImportError:
    PARSERS.append(
        pytest.param(
            None, id="re2", marks=pytest.mark.skip(reason="re2 parser not available")
        )
    )
else:
    PARSERS.append(pytest.param(Parser(re2.Resolver(data)), id="re2"))

try:
    from ua_parser import regex
except ImportError:
    PARSERS.append(
        pytest.param(
            None,
            id="regex",
            marks=pytest.mark.skip(reason="regex parser not available"),
        )
    )
else:
    PARSERS.append(pytest.param(Parser(regex.Resolver(data)), id="regex"))

UA_FIELDS = {f.name for f in dataclasses.fields(UserAgent)}


@pytest.mark.parametrize("parser", PARSERS)
@pytest.mark.parametrize(
    "test_file",
    [
        CORE_DIR / "tests" / "test_ua.yaml",
        CORE_DIR / "test_resources" / "firefox_user_agent_strings.yaml",
        CORE_DIR / "test_resources" / "pgts_browser_list.yaml",
        CORE_DIR / "test_resources" / "opera_mini_user_agent_strings.yaml",
        CORE_DIR / "test_resources" / "podcasting_user_agent_strings.yaml",
    ],
    ids=attrgetter("stem"),
)
def test_ua(parser, test_file):
    with test_file.open("rb") as f:
        contents = load(f, Loader=SafeLoader)

    for test_case in contents["test_cases"]:
        res = {k: v for k, v in test_case.items() if k in UA_FIELDS}
        # there seems to be broken test cases which have a patch_minor
        # of null where it's not, as well as the reverse, so we can't
        # test patch_minor (ua-parser/uap-core#562)
        res.pop("patch_minor", None)
        r = parser.parse_user_agent(test_case["user_agent_string"]) or UserAgent()
        assert dataclasses.asdict(r).items() >= res.items()


OS_FIELDS = {f.name for f in dataclasses.fields(OS)}


@pytest.mark.parametrize("parser", PARSERS)
@pytest.mark.parametrize(
    "test_file",
    [
        CORE_DIR / "tests" / "test_os.yaml",
        CORE_DIR / "test_resources" / "additional_os_tests.yaml",
    ],
    ids=attrgetter("stem"),
)
def test_os(parser, test_file):
    with test_file.open("rb") as f:
        contents = load(f, Loader=SafeLoader)

    for test_case in contents["test_cases"]:
        res = {k: v for k, v in test_case.items() if k in OS_FIELDS}
        r = parser.parse_os(test_case["user_agent_string"]) or OS()
        assert dataclasses.asdict(r) == res


DEVICE_FIELDS = {f.name for f in dataclasses.fields(Device)}


@pytest.mark.parametrize("parser", PARSERS)
@pytest.mark.parametrize(
    "test_file",
    [
        CORE_DIR / "tests" / "test_device.yaml",
    ],
    ids=attrgetter("stem"),
)
def test_devices(parser, test_file):
    with test_file.open("rb") as f:
        contents = load(f, Loader=SafeLoader)

    for test_case in contents["test_cases"]:
        res = {k: v for k, v in test_case.items() if k in DEVICE_FIELDS}
        r = parser.parse_device(test_case["user_agent_string"]) or Device()
        assert dataclasses.asdict(r) == res


def test_results():
    p = Parser(BasicResolver(([UserAgentMatcher("(x)")], [], [])))

    assert p.parse_user_agent("x") == UserAgent("x")
    assert p.parse_user_agent("y") is None

    assert p.parse("x") == Result(
        user_agent=UserAgent("x"),
        os=None,
        device=None,
        string="x",
    )
    assert p.parse("y") == Result(
        user_agent=None,
        os=None,
        device=None,
        string="y",
    )



import platform
import re
from typing import Match, Optional

IS_GRAAL: bool = platform.python_implementation() == "GraalVM"


def get(m: Match[str], idx: int) -> Optional[str]:
    return (m[idx] or None) if 0 < idx <= m.re.groups else None


def replacer(repl: str, m: Match[str]) -> Optional[str]:
    """The replacement rules are frustratingly subtle and innimical to
    standard python fallback semantics:

    - if there is a non-null replacement pattern, then it must be used with
      match groups as template parameters (at indices 1+)
      - the result is stripped
      - if it is an empty string, then it's replaced by a null
    - otherwise fallback to a (possibly optional) match group
    - or null (device brand has no fallback)

    Replacement rules only apply to OS and Device matchers, the UA
    matcher has bespoke replacement semantics for the family (just
    $1), and no replacement for the other fields, either there is a
    static replacement or it falls back to the corresponding
    (optional) match group.

    """
    if not repl:
        return None

    return re.sub(r"\$(\d)", lambda n: get(m, int(n[1])) or "", repl).strip() or None


REPETITION_PATTERN = re.compile(r"\{(0|1)\s*,\s*\d{3,}\}")
CLASS_PATTERN = re.compile(
    r"""
\[[^]]*\\(d|w)[^]]*\]
|
\\(d|w)
""",
    re.VERBOSE,
)


def class_replacer(m: re.Match[str]) -> str:
    d, w = ("0-9", "A-Za-z0-9_") if m[1] else ("[0-9]", "[A-Za-z0-9_]")
    return m[0].replace(r"\d", d).replace(r"\w", w)


def fa_simplifier(pattern: str) -> str:
    """uap-core makes significant use of large bounded repetitions, to
    mitigate catastrophic backtracking.

    However this explodes the number of states (and thus graph size)
    for finite automaton engines, which significantly increases their
    memory use, and for those which use JITs it can exceed the JIT
    threshold and force fallback to a slower engine (seems to be the
    case for graal's TRegex).
    """
    pattern = REPETITION_PATTERN.sub(lambda m: "*" if m[1] == "0" else "+", pattern)
    return CLASS_PATTERN.sub(class_replacer, pattern)



import pytest  # type: ignore

from ua_parser import Domain, PartialResult

re2 = pytest.importorskip("ua_parser.re2")


def test_empty(capfd: pytest.CaptureFixture[str]) -> None:
    r = re2.Resolver(([], [], []))
    assert r("", Domain.ALL) == PartialResult(Domain.ALL, None, None, None, "")
    out, err = capfd.readouterr()
    assert out == ""
    assert err == ""



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os
import sys

sys.path.insert(0, os.path.normpath(os.path.join(__file__, "../..", "src")))
# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "UA Parser"
copyright = "2024, UA Parser Project"
author = "UA Parser Project"

version = "1.0"
release = "1.0"

rst_epilog = """
.. |pyyaml| replace:: ``PyYaml``
.. |re2| replace:: ``google-re2``
.. |regex| replace:: ``regex``

.. _pyyaml: https://pyyaml.org
.. _re2: https://pypi.org/project/google-re2
.. _regex: https://pypi.org/project/ua-parser-rs
"""

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

language = "en"

html_theme = "alabaster"

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}



from collections.abc import Iterable
from typing import Literal, Protocol

UAParser = tuple[
    str,
    str | None,
    str | None,
    str | None,
    str | None,
    str | None,
]

class UserAgent(Protocol):
    family: str
    major: str | None
    minor: str | None
    patch: str | None
    patch_minor: str | None

class UserAgentExtractor:
    def __init__(self, it: Iterable[UAParser], /) -> None: ...
    def extract(self, s: str, /) -> UserAgent | None: ...

OSParser = tuple[
    str,
    str | None,
    str | None,
    str | None,
    str | None,
    str | None,
]

class OS(Protocol):
    family: str
    major: str | None
    minor: str | None
    patch: str | None
    patch_minor: str | None

class OSExtractor:
    def __init__(self, it: Iterable[OSParser], /) -> None: ...
    def extract(self, s: str, /) -> OS | None: ...

DeviceParser = tuple[
    str,
    Literal["i"] | None,
    str | None,
    str | None,
    str | None,
]

class Device(Protocol):
    family: str
    brand: str | None
    model: str | None

class DeviceExtractor:
    def __init__(self, it: Iterable[DeviceParser], /) -> None: ...
    def extract(self, s: str, /) -> Device | None: ...



from __future__ import annotations

import abc
import dataclasses
import threading
from collections import OrderedDict, deque
from contextvars import ContextVar
from typing import (
    Callable,
    Deque,
    Dict,
    Optional,
    Protocol,
    Union,
)

from .core import Domain, PartialResult, Resolver

__all__ = [
    "Cache",
    "CachingResolver",
    "Lru",
    "S3Fifo",
    "Sieve",
]


class Cache(Protocol):
    """Cache()

    Cache abstract protocol. The :class:`CachingResolver` will look
    values up, merge what was returned (possibly nothing) with what it
    got from its actual parser, and *re-set the result*.
    """

    @abc.abstractmethod
    def __setitem__(self, key: str, value: PartialResult) -> None:
        """Adds or replace ``value`` to the cache at key ``key``."""
        ...

    @abc.abstractmethod
    def __getitem__(self, key: str) -> Optional[PartialResult]:
        """Returns a partial result for ``key`` if there is any."""
        ...


class Lru:
    """Cache following a least-recently used replacement policy: when
    there is no more room in the cache, whichever entry was last seen
    the least recently is removed.

    Simple LRUs are generally outdated and to avoid as they have
    relatively low hit rates for modern caches (at lower sizes). The
    main use case here is if the workload can lead to the cache being
    full of popular items then all of them being replaced at once:
    :class:`S3Fifo` and :class:`Sieve` are FIFO-based caches and have
    worst-case O(n) eviction.
    """

    def __init__(self, maxsize: int):
        self.maxsize = maxsize
        self.cache: OrderedDict[str, PartialResult] = OrderedDict()
        self.lock = threading.Lock()

    def __getitem__(self, key: str) -> Optional[PartialResult]:
        with self.lock:
            e = self.cache.get(key)
            if e:
                self.cache.move_to_end(key)
            return e

    def __setitem__(self, key: str, value: PartialResult) -> None:
        with self.lock:
            if len(self.cache) >= self.maxsize and key not in self.cache:
                self.cache.popitem(last=False)
            self.cache[key] = value


@dataclasses.dataclass
class CacheEntry:
    __slots__ = ["freq", "key", "value"]
    key: str
    value: PartialResult
    freq: int


class S3Fifo:
    """FIFO-based quick-demotion lazy-promotion cache [S3-FIFO]_.

    Experimentally provides excellent hit rate at lower cache sizes,
    for a relatively simple and efficient implementation. Notably
    excellent at handling "one hit wonders", aka entries seen only
    once during a work-set (or reasonable work window).
    """

    def __init__(self, maxsize: int):
        self.maxsize = maxsize
        self.index: Dict[str, Union[CacheEntry, str]] = {}
        self.small_target = max(1, int(maxsize / 10))
        self.small: Deque[CacheEntry] = deque()
        self.main_target = maxsize - self.small_target
        self.main: Deque[CacheEntry] = deque()
        self.ghost: Deque[str] = deque()
        self.lock = threading.Lock()

    def __getitem__(self, key: str) -> Optional[PartialResult]:
        if (e := self.index.get(key)) and type(e) is CacheEntry:
            # small race here, we could miss an increment
            e.freq = min(e.freq + 1, 3)
            return e.value

        return None

    def __setitem__(self, key: str, r: PartialResult) -> None:
        with self.lock:
            if (e := self.index.get(key)) and type(e) is CacheEntry:
                e.value = r
                return

            if len(self.small) + len(self.main) >= self.maxsize:
                # if main is not overcapacity, resize small
                if len(self.main) < self.main_target:
                    self._evict_small()
                # evict_small could have moved every entry to main, in
                # which case we now need to evict from main
                if len(self.small) + len(self.main) >= self.maxsize:
                    self._evict_main()

            entry = CacheEntry(key, r, 0)
            if type(self.index.get(key)) is str:
                self.main.appendleft(entry)
            else:
                self.small.appendleft(entry)
            self.index[key] = entry

    def _evict_main(self) -> None:
        while True:
            e = self.main.pop()
            if e.freq:
                e.freq -= 1
                self.main.appendleft(e)
            else:
                del self.index[e.key]
                return

    def _evict_small(self) -> None:
        while self.small:
            e = self.small.pop()
            if e.freq:
                e.freq = 0
                self.main.appendleft(e)
            else:
                g = self.index[e.key] = e.key
                self.ghost.appendleft(g)
                while len(self.ghost) > self.main_target:
                    g = self.ghost.pop()
                    if self.index.get(g) is g:
                        del self.index[g]
                return


@dataclasses.dataclass
class SieveNode:
    __slots__ = ("key", "next", "value", "visited")
    key: str
    value: PartialResult
    visited: bool
    next: Optional[SieveNode]


class Sieve:
    """FIFO-based quick-demotion cache [SIEVE]_.

    Simpler FIFO-based cache, cousin of :class:`S3Fifo`.
    Experimentally slightly lower hit rates than :class:`S3Fifo` (if
    way superior to LRU still), but much more compact (~50% lower
    memory overhead at larger cache sizes, up to 100% at very small
    cache sizes).

    Can be an interesting candidate when trying to save on memory,
    although the contained entries will generally be much larger than
    the cache itself.
    """

    def __init__(self, maxsize: int) -> None:
        self.maxsize = maxsize
        self.cache: Dict[str, SieveNode] = {}
        self.head: Optional[SieveNode] = None
        self.tail: Optional[SieveNode] = None
        self.hand: Optional[SieveNode] = None
        self.prev: Optional[SieveNode] = None
        self.lock = threading.Lock()

    def __getitem__(self, key: str) -> Optional[PartialResult]:
        if entry := self.cache.get(key):
            entry.visited = True
            return entry.value

        return None

    def __setitem__(self, key: str, value: PartialResult) -> None:
        with self.lock:
            if e := self.cache.get(key):
                e.value = value
                return

            if len(self.cache) >= self.maxsize:
                self._evict()

            node = self.cache[key] = SieveNode(key, value, False, None)
            if self.head:
                self.head.next = node
            self.head = node
            if self.tail is None:
                self.tail = node

    def _evict(self) -> None:
        obj: Optional[SieveNode]
        if self.hand:
            obj, pobj = self.hand, self.prev
        else:
            obj, pobj = self.tail, None

        while obj and obj.visited:
            obj.visited = False
            if obj.next:
                obj, pobj = obj.next, obj
            else:
                obj, pobj = self.tail, None

        if not obj:
            return

        self.hand = obj.next
        self.prev = pobj

        del self.cache[obj.key]
        if not obj.next:
            self.head = pobj

        if pobj:
            pobj.next = obj.next
        else:
            self.tail = obj.next


class Local:
    """Thread local cache decorator. Takes a cache factory and lazily
    instantiates a cache for each thread it's accessed from.

    This means the cache capacity and memory consumption is
    figuratively multiplied by however many threads the cache is used
    from, but those threads don't share their caching, and thus don't
    contend on cache use.

    """

    def __init__(self, factory: Callable[[], Cache]) -> None:
        self.cv: ContextVar[Cache] = ContextVar("local-cache")
        self.factory = factory

    @property
    def cache(self) -> Cache:
        c = self.cv.get(None)
        if c is None:
            c = self.factory()
            self.cv.set(c)
        return c

    def __getitem__(self, key: str) -> Optional[PartialResult]:
        return self.cache[key]

    def __setitem__(self, key: str, value: PartialResult) -> None:
        self.cache[key] = value


class CachingResolver:
    """A wrapper resolver which takes an underlying concrete
    :class:`Cache` for the actual caching and cache strategy.

    This resolver only interacts with the :class:`Cache` and delegates
    to the wrapped resolver in case of lookup failure.

    :class:`CachingParser` will set entries back in the cache when
    filling them up, it does not update results in place (and can't
    really, they're immutable).

    """

    def __init__(self, resolver: Resolver, cache: Cache):
        self.parser: Resolver = resolver
        self.cache: Cache = cache

    def __call__(self, ua: str, domains: Domain, /) -> PartialResult:
        entry = self.cache[ua]
        if entry:
            if domains in entry.domains:
                return entry

            domains &= ~entry.domains

        r = self.parser(ua, domains)
        if entry:
            r = PartialResult(
                string=ua,
                domains=entry.domains | r.domains,
                user_agent=entry.user_agent or r.user_agent,
                os=entry.os or r.os,
                device=entry.device or r.device,
            )
        self.cache[ua] = r
        return r
